#include <conio.h>
#include "sin.h"
#include "cos.h"

#define PERSPECT      0x147
#define CORRECT       54613L
#define NUMPOINTS     200
#define NUMTRIANGLES  400
#define D1POINTS      20
#define D2POINTS      10
#define D1ANGLE       18
#define D2ANGLE       36

extern "C" TRIANGLE(unsigned x1, unsigned y1,
                    unsigned x2, unsigned y2,
                    unsigned x3, unsigned y3,
                    unsigned color, unsigned page);

extern "C" long SQRT(long arg);

unsigned page = 0;

class TPoint
{
  public:
    long x, y, z;
    TPoint() {}
    TPoint(long x, long y, long z)
    {
      TPoint::x = x;
      TPoint::y = y;
      TPoint::z = z;
    }
    void CorrectAspect()
    {
      y = int((y * CORRECT) >> 16);
    }
};

TPoint RotateX(TPoint p, unsigned angle)
{
  return TPoint(p.x,
                (p.y * cos_table[angle] - p.z * sin_table[angle]) >> 16,
                (p.y * sin_table[angle] + p.z * cos_table[angle]) >> 16);
}

TPoint RotateY(TPoint p, unsigned angle)
{
  return TPoint((p.x * cos_table[angle] - p.z * sin_table[angle]) >> 16,
                p.y,
                (p.x * sin_table[angle] + p.z * cos_table[angle]) >> 16);
}

TPoint RotateZ(TPoint p, unsigned angle)
{
  return TPoint((p.x * cos_table[angle] - p.y * sin_table[angle]) >> 16,
                (p.x * sin_table[angle] + p.y * cos_table[angle]) >> 16,
                p.z);
}

class TNormal
{
    long x, y, z;
  public:
    TNormal() {}
    TNormal(long x, long y, long z)
    {
      TNormal::x = x;
      TNormal::y = y;
      TNormal::z = z;
    }
    TNormal(TPoint p1, TPoint p2, TPoint p3)
    {
      long register t;
      TNormal n1(p2.x - p1.x, p2.y - p1.y, p2.z - p1.z),
              n2(p3.x - p1.x, p3.y - p1.y, p3.z - p1.z);

      x = n1.y * n2.z - n1.z * n2.y;
      y = n1.z * n2.x - n1.x * n2.z;
      z = n1.x * n2.y - n1.y * n2.x;

      t = (long)SQRT(x * x + y * y + z * z);

      x = (x << 16) / t;
      y = (y << 16) / t;
      z = (z << 16) / t;
    }
    long Z()
    {
      return z;
    }
};

void Convert3D(TPoint p, int &xx, int &yy)
{
  long register t = (long)p.z * PERSPECT;
  long register t_x = p.x, t_y = p.y;

  if (p.z > 0)
  {
    t_x *= t + 0x10000L;
    t_y *= t + 0x10000L;
    xx = int(t_x >> 16);
    yy = int(t_y >> 16);
  }
  else
  {
    t_x <<= 16;
    t_y <<= 16;
    t_x /= 0x10000L - t;
    t_y /= 0x10000L - t;
    xx = (int)t_x;
    yy = (int)t_y;
  }
}

void Triangle2D(TPoint p1, TPoint p2, TPoint p3,
                unsigned x, unsigned y, unsigned color)
{
  int t_x1, t_y1, t_x2, t_y2, t_x3, t_y3;

  Convert3D(p1, t_x1, t_y1);
  Convert3D(p2, t_x2, t_y2);
  Convert3D(p3, t_x3, t_y3);

  TRIANGLE(t_x1 + x, t_y1 + y,
           t_x2 + x, t_y2 + y,
           t_x3 + x, t_y3 + y,
           color, page);
}

class TTriangle {
    TPoint m1, m2, m3;
    int light, sum;
    TNormal n;
  public:
    TTriangle(): n(), m1(), m2(), m3() {}
    TTriangle(TPoint p1, TPoint p2, TPoint p3): n(p1, p2, p3), m1(p1), m2(p2), m3(p3)
    {
      light = int(((n.Z() << 6) - 1) >> 16);
      sum = int(m1.z + m2.z + m3.z);
    }
    int Visible()
    {
      return (n.Z() > 0);
    }
    void Draw(unsigned color, unsigned x, unsigned y)
    {
      Triangle2D(m1, m2, m3, x, y, color);
    }
    int Light()
    {
      return light;
    }
    int SetLight(int light)
    {
      if (light > 0)
        TTriangle::light = light;
    }
    long ZSum()
    {
      return sum;
    }
};

class TTorus {
    TPoint p[NUMPOINTS], p2[NUMPOINTS];
    int v[NUMTRIANGLES];
    TTriangle t[NUMTRIANGLES];
    void InsertItem(int index, int item);
    int s;
  public:
    TTorus(int d1, int d2);
    void Draw(unsigned, unsigned, unsigned, unsigned, unsigned, unsigned);
};

TTorus::TTorus(int d1, int d2)
{
  for (int i = 0; i < D1POINTS; i++)
  {
    for (int j = 0; j < D2POINTS; j++)
      p[i * D2POINTS + j] = TPoint(RotateY(TPoint(((cos_table[j * D2ANGLE] * (long)d2) >> 16) + d1,
                                                  (sin_table[j * D2ANGLE] * (long)d2) >> 16, 0),
                                           i * D1ANGLE));
  }
}

void TTorus::InsertItem(int index, int item)
{
  for (int i = s; i > index; i--)
    v[i] = v[i - 1];
  v[index] = item;
  s++;
}

void TTorus::Draw(unsigned angle_x, unsigned angle_y,
                  unsigned angle_z, unsigned x,
                  unsigned y, unsigned color)
{
  int i = 0, offs, j, a, b, c;

  for (; i < NUMPOINTS; i++)
  {
    p2[i] = TPoint(RotateX(RotateY(RotateZ(p[i], angle_z), angle_y), angle_x));
    p2[i].CorrectAspect();
  }

  for (i = 0; i < D1POINTS; i++)
  {
    for (j = 0; j < D2POINTS; j++)
    {
      a = i + 1;
      if ((i + 1) == D1POINTS) a = 0;
      a = (a << 1) + (a << 3);
      b = (i << 1) + (i << 3);
      c = j + 1;
      if ((j + 1) == D2POINTS) c = 0;
      t[b + j] = TTriangle(p2[b + j], p2[b + c], p2[a + c]);
      t[b + j + NUMPOINTS] = TTriangle(p2[b + j], p2[a + c], p2[a + j]);
      t[b + j + NUMPOINTS].SetLight(t[b + j].Light());
    }
  }

  s = 0;

  for (i = 0; i < NUMTRIANGLES; i++)
  {
    if (t[i].Visible())
    {
      for (j = 0; j < s; j++)
      {
        if (t[i].ZSum() <= t[v[j]].ZSum())
        {
          InsertItem(j, i);
          break;
        }
      }
      if (j == s) InsertItem(s, i);
    }
  }

  for (i = 0; i < s; i++)
  {
    offs = t[v[i]].Light();
    t[v[i]].Draw((color << 6) + offs, x, y);
  }
}

class TFlyer {
    int x, y;
    int a_x, a_y, a_z;
    unsigned size, color;
    int dx, dy, d_ax, d_ay, d_az;
    TTorus t;
  public:
    TFlyer(unsigned size, int points, int dx, int dy, int d_ax, int d_ay, int d_az,
           unsigned color):
          t(size, points)
    {
      TFlyer::size = size;
      TFlyer::color = color;
      TFlyer::dx = dx;
      TFlyer::dy = dy;
      TFlyer::d_ax = d_ax;
      TFlyer::d_ay = d_ay;
      TFlyer::d_az = d_az;
      a_x = a_y = a_z = 0;
      x = 160;
      y = 100;
    }
    void Move()
    {
      t.Draw(a_x, a_y, a_z, x, y, color);

      x += dx;
      y += dy;

      a_x += d_ax;
      a_y += d_ay;
      a_z += d_az;

      if ((x > 320 - size - size / 2) || (x < size + size / 2)) dx = -dx;
      if ((y > 200 - size - size / 2) || (y < size + size / 2)) dy = -dy;

      if (a_x < 0) a_x = 360 + a_x;
      if (a_x > 359) a_x = 0;
      if (a_y < 0) a_y = 360 + a_y;
      if (a_y > 359) a_y = 0;
      if (a_z < 0) a_z = 360 + a_z;
      if (a_z > 359) a_z = 0;
    }
};

void SetGraphMode()
{
  asm  mov    ax,  13h
  asm  int    10h
  asm  mov    dx,  3C4h
  asm  mov    al,  4
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  or     al,  4
  asm  and    al,  0F7h
  asm  out    dx,  al
  asm  mov    dx,  3D4h
  asm  mov    al,  14h
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  and    al,  0BFh
  asm  out    dx,  al
  asm  dec    dx
  asm  mov    al,  17h
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  or     al,  40h
  asm  out    dx,  al
  asm  push   0A000h
  asm  pop    es
  asm  xor    di,  di
  asm  xor    ax,  ax
  asm  mov    cx,  8000h
  asm  cld
  asm  rep    stosw
}

void SetTextMode()
{
  asm  mov    ax,  03h
  asm  int    10h
}

void ClearPage(unsigned page)
{
  asm  push   0A000h
  asm  pop    es
  asm  mov    di,  page
  asm  xor    ax,  ax
  asm  mov    cx,  8192
  asm  mov    dx,  3C4h
  asm  mov    ax,  0F02h
  asm  out    dx,  ax
  asm  cld
  asm  mov    cx,  8192
  asm  xor    ax,  ax
  asm  rep    stosw
}

void TilePage1(unsigned page, char color)
{
  asm  push   0A000h
  asm  pop    es
  asm  mov    di,  page
  asm  mov    dx,  3C4h
  asm  mov    cx,  50
  asm  cld
l:asm  push   cx
  asm  mov    bl,  color
  asm  add    bl,  cl
  asm  mov    bh,  bl
  asm  mov    ax,  0F02h
  asm  out    dx,  ax
  asm  mov    cx,  40
  asm  mov    ax,  bx
  asm  rep    stosw
  asm  push   di
  asm  mov    ax,  0E02h
  asm  out    dx,  ax
  asm  mov    cx,  40
  asm  xor    ax,  ax
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  pop    di
  asm  mov    ax,  0102h
  asm  out    dx,  ax
  asm  mov    cx,  40
  asm  mov    ax,  bx
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  pop    cx
  asm  loop   l
}

void TilePage2(unsigned page, char color)
{
  asm  push   0A000h
  asm  pop    es
  asm  mov    di,  page
  asm  mov    dx,  3C4h
  asm  mov    ax,  0F02h
  asm  out    dx,  ax
  asm  mov    cx,  50
  asm  mov    bl,  color
  asm  xor    bh,  bh
  asm  cld
l:asm  push   cx
  asm  mov    ax,  bx
  asm  add    al,  cl
  asm  test   cl,  1
  asm  jz     m
  asm  xchg   ah,  al
m:asm  mov    cx,  40
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  pop    cx
  asm  loop   l
}

void SetPage(unsigned page)
{
  asm  mov    bx,  page
  asm  mov    dx,  3D4h
  asm  mov    al,  0Ch
  asm  mov    ah,  bh
  asm  out    dx,  ax
  asm  mov    al,  0Dh
  asm  xor    ah,  ah
  asm  out    dx,  ax
}

void Vertret()
{
    asm  mov    dx,  03DAh
l1: asm  in     al,  dx
    asm  test   al,  8
    asm  jz     l1
l2: asm  in     al,  dx
    asm  test   al,  8
    asm  jnz    l2
}

void SetColor(char color, char r, char g, char b)
{
  asm  mov    dx,  3C8h
  asm  mov    al,  color
  asm  out    dx,  al
  asm  inc    dx
  asm  mov    al,  r
  asm  out    dx,  al
  asm  mov    al,  g
  asm  out    dx,  al
  asm  mov    al,  b
  asm  out    dx,  al
}

void SetPalette()
{
  for (int i = 0; i < 64; i++)
  {
    SetColor(i, 0, 0, i);
    SetColor(i + 64, 0, i, 0);
    SetColor(i + 128, i, 0, 0);
    SetColor(i + 192, i, i, 0);
  }
}

void Message()
{
  cprintf("3D transformations - by HackMaster\n");
}

void main()
{
  SetGraphMode();
  SetPalette();
  TFlyer f1(60, 30, 0, 0, 0, 2, 2, 2);

  do
  {
    SetPage(page ^ 0x4000);
    Vertret();
    TilePage2(page, 0x00);
    f1.Move();
    page ^= 0x4000;
  } while (!kbhit());

  SetTextMode();
  Message();
}
